import sys
import typer

from chris.cube.pipeline import Pipeline
from chris.cube.plugin_instance import PluginInstance


def run_pipeline_with_progress(chris_pipeline: Pipeline, plugin_instance: PluginInstance):
    """
    Helper to execute a pipeline with a progress bar.
    """
    plugin_root = chris_pipeline.get_root()
    with typer.progressbar(plugin_root.run(plugin_instance.id),
                           length=len(plugin_root), label='Scheduling pipeline',
                           file=sys.stderr) as proto_pipeline:
        for _ in proto_pipeline:
            pass



from dataclasses import dataclass

from chris.types import CUBEUrl, FeedId, ISOFormatDateString, CUBEUsername, FilesUrl

from chris.helpers.collection import collection_helper
from chris.cube.resource.templated_resource import ResourceWithTemplate

from typing import List


@dataclass(frozen=True)
class Feed(ResourceWithTemplate):
    """
    A *feed* in *ChRIS* is a DAG of *plugin instances*.
    """
    id: FeedId
    creation_date: ISOFormatDateString
    modification_date: ISOFormatDateString
    name: str
    creator_username: CUBEUsername
    created_jobs: int
    waiting_jobs: int
    scheduled_jobs: int
    started_jobs: int
    registering_jobs: int
    finished_jobs: int
    errored_jobs: int
    cancelled_jobs: int
    owner: List[CUBEUrl]
    note: CUBEUrl
    tags: CUBEUrl
    taggings: CUBEUrl
    comments: CUBEUrl
    files: FilesUrl
    plugin_instances: CUBEUrl

    def set_name(self, name: str) -> dict:
        return self.__put(
            url=self.url,
            data={'name': name}
        )

    def set_description(self, description: str) -> dict:
        return self.__put(
            url=self.note,
            data={
                'title': 'Description',
                'content': description
            }
        )

    def __put(self, url: CUBEUrl, data: dict) -> dict:
        payload = collection_helper(data)
        res = self.s.put(url, json=payload)
        res.raise_for_status()
        return res.json()



def collection_helper(obj: dict = None) -> dict:
    if not obj:
        obj = {}
    return {
        'template': {
            'data': [{'name': k, 'value': v} for k, v in obj.items()]
        }
    }



import abc
from typing import Optional

import logging

logger = logging.getLogger(__name__)


use_keyring = True
"""
``use_keyring`` is a :class:`bool` which denotes whether it is possible to use
:class:`KeyringSecretStore`.
"""

try:
    import keyring
    logger.info('Using keyring')
except ModuleNotFoundError:
    use_keyring = False
    logger.info('Using plaintext')


class AbstractSecretStore(abc.ABC):
    """
    A key-value data structure which hopefully stores its values securely.

    Implementations of ``AbstractSecretStore`` are given a :class:`dict` which
    the ``AbstractSecretStore`` may optionally use as a persistent key-value
    store for arbitrary metadata. However, it is the ``AbstractSecretStore``'s
    client's responsibility to read and write that :class:`dict` to disk, thus
    it is considered insecure.
    """

    def __init__(self, context: dict):
        """
        :param context: application settings which are expected to be saved
                        after the invocation of this object's methods
        """
        self.context = context

    @abc.abstractmethod
    def get(self, address: str) -> Optional[str]:
        ...

    @abc.abstractmethod
    def set(self, address: str, token: str):
        ...

    @abc.abstractmethod
    def clear(self, address: str):
        ...


class PlaintextSecretStore(AbstractSecretStore):
    """
    Insecure token storage in plaintext.
    This implementation should only be used when the keyring is not available.
    """
    def __init__(self, context: dict):
        super().__init__(context)
        if 'secrets' not in context:
            context['secrets'] = {}

    def get(self, address: str) -> Optional[str]:
        if address not in self.context['secrets']:
            return None
        return self.context['secrets'][address]

    def set(self, address: str, token: str):
        self.context['secrets'][address] = token

    def clear(self, address: str):
        del self.context['secrets'][address]


class KeyringSecretStore(AbstractSecretStore):
    """
    Secure token storage using the host desktop environment's login keyring.
    """

    __SPACE = 'org.chrisproject.caw'

    def get(self, address: str) -> Optional[str]:
        return keyring.get_password(self.__SPACE, address)

    def set(self, address: str, token: str):
        keyring.set_password(self.__SPACE, address, token)

    def clear(self, address: str):
        keyring.delete_password(self.__SPACE, address)



from typing import Optional
import typer

from chris.types import CUBEAddress, CUBEUsername, CUBEPassword
import caw
from caw.commands.store import app, build_client
from caw.constants import DEFAULT_ADDRESS, DEFAULT_USERNAME, DEFAULT_PASSWORD


def show_version(value: bool):
    """
    Print version.
    """
    if not value:
        return
    typer.echo(f'{caw.pkg.metadata["name"]} {caw.__version__}')
    raise typer.Exit()


# noinspection PyUnusedLocal
@app.callback()
def entry(
        address: str = typer.Option(DEFAULT_ADDRESS, '--address', '-a', envvar='CHRIS_URL'),
        username: Optional[str] = typer.Option(
            None, '--username', '-u', envvar='CHRIS_USERNAME',
            help='Username of your ChRIS user account.'),
        password: Optional[str] = typer.Option(
            None, '--password', '-p', envvar='CHRIS_PASSWORD',
            help='Password of your ChRIS user account. '
            'If neither username nor password are specified, then the default '
            f'account "{DEFAULT_USERNAME}:{DEFAULT_PASSWORD}" is used.'),
        version: Optional[bool] = typer.Option(
            None, '--version', '-V', callback=show_version, is_eager=True, help='Print version.')
):
    """
    A command line ChRIS client for pipeline execution and data management.
    """
    build_client.address = CUBEAddress(address)
    build_client.username = CUBEUsername(username)
    build_client.password = CUBEPassword(password)



import os
import requests
from datetime import datetime
from chris.errors import PaginationNotImplementedException
from typing import Optional, Union, Iterator
from collections.abc import Iterable
from pathlib import Path
import logging






class UploadedFile(ConnectedResource):
    """
    Represents a file resource in CUBE.

    TODO: rename to RegisteredFile
    """
    def __init__(self, creation_date: str, file_resource: str, fname: str, fsize: int, id: int,
                 url: str, session: requests.Session,
                 owner: Optional[str] = None,
                 feed_id: Optional[int] = None, plugin_inst: Optional[str] = None,
                 plugin_inst_id: Optional[int] = None,

                 PatientID: Optional[str] = None,
                 PatientName: Optional[str] = None,
                 PatientBirthDate: Optional[str] = None,
                 PatientAge: Optional[float] = None,  # float or int?
                 PatientSex: Optional[str] = None,
                 StudyDate: Optional[str] = None,
                 Modality: Optional[str] = None,
                 ProtocolName: Optional[str] = None,
                 StudyInstanceUID: Optional[str] = None,
                 StudyDescription: Optional[str] = None,
                 SeriesInstanceUID: Optional[str] = None,
                 SeriesDescription: Optional[str] = None,
                 pacs_identifier: Optional[str] = None):
        super().__init__(url, session)
        self.creation_date = datetime.fromisoformat(creation_date)
        self.file_resource = file_resource
        self.fname = fname
        self.fsize = fsize
        self.id = id
        self.ownder = owner
        self.feed_id = feed_id
        self.plugin_inst = plugin_inst
        self.plugin_inst_id = plugin_inst_id

        self.PatientID = PatientID
        self.PatientName = PatientName
        self.PatientBirthDate = PatientBirthDate
        self.PatientAge: Optional[float] = PatientAge
        self.PatientSex = PatientSex
        self.StudyDate = StudyDate
        self.Modality = Modality
        self.ProtocolName = ProtocolName
        self.StudyInstanceUID = StudyInstanceUID
        self.StudyDescription = StudyDescription
        self.SeriesInstanceUID = SeriesInstanceUID
        self.SeriesDescription = SeriesDescription
        self.pacs_identifier = pacs_identifier

    def download(self, destination: Union[Path, str], chunk_size=8192):
        with self._s.get(self.file_resource, stream=True, headers={'Accept': None}) as r:
            r.raise_for_status()
            with open(destination, 'wb') as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    f.write(chunk)


class InvalidFilesResourceUrlException(Exception):
    pass


class UploadedFiles(ConnectedResource, Iterable):
    """
    Lazy iterable over paginated response.
    """
    __logger = logging.getLogger('UploadedFiles')

    def __init__(self, url: str, session: requests.Session):
        super().__init__(url, session)
        if 'limit=' not in self.url:
            self.url += f"{'&' if '?' in self.url else '?'}limit={PAGINATION_LIMIT}"

        try:
            self._initial_data = self._do_get(self.url)
        except requests.exceptions.HTTPError as e:
            raise InvalidFilesResourceUrlException(f'{e.response.status_code} error getting {self.url}')
        # check given URL is a files collection resource
        if 'count' not in self._initial_data:
            raise InvalidFilesResourceUrlException(f'{self.url} does not look like a files collection resource.')
        if self._initial_data['count'] > 0:
            try:
                UploadedFile(session=self._s, **self._initial_data['results'][0])
            except KeyError:
                raise InvalidFilesResourceUrlException(f'{self.url} is not a files collection resource.')

    def __iter__(self) -> Iterator[UploadedFile]:
        data = self._initial_data  # first page
        if data['previous'] is not None:
            self.__logger.warning('%s is not the first page.', self.url)

        while data['next']:
            for fdata in data['results']:
                yield UploadedFile(**fdata, session=self._s)
            self.__logger.debug('next page: %s', data['next'])
            data = self._do_get(data['next'])  # next page
        for fdata in data['results']:  # last page
            yield UploadedFile(**fdata, session=self._s)

    def __len__(self):
        return self._initial_data['count']

    def _do_get(self, url):
        self.__logger.info('getting %s', url)
        res = self._s.get(url)
        res.raise_for_status()
        return res.json()



from dataclasses import dataclass
from typing import Optional
from datetime import datetime

from chris.cube.resource.cube_resource import CUBEResource
from chris.cube.feed import Feed
from chris.types import (
    CUBEUsername, CUBEUrl, PluginName, PluginVersion, PluginType,
    PluginInstanceId, FeedId, PluginId, ComputeResourceName,
    SwiftPath, ISOFormatDateString, PluginInstanceStatus, CUBEErrorCode
)


# It'd be better to use inheritance instead of optionals
@dataclass(frozen=True)
class PluginInstance(CUBEResource):
    """
    A *plugin instance* in _ChRIS_ is a computing job, i.e. an attempt to run
    a computation (a non-interactive command-line app) to produce data.
    """
    id: Optional[PluginInstanceId]
    title: str
    compute_resource_name: ComputeResourceName
    plugin_id: PluginId
    plugin_name: PluginName
    plugin_version: PluginVersion
    plugin_type: PluginType

    pipeline_inst: Optional[int]
    feed_id: FeedId
    start_date: ISOFormatDateString
    end_date: ISOFormatDateString
    output_path: SwiftPath

    status: PluginInstanceStatus

    summary: str
    raw: str
    owner_username: CUBEUsername
    cpu_limit: int
    memory_limit: int
    number_of_workers: int
    gpu_limit: int
    error_code: CUBEErrorCode

    previous: CUBEUrl
    feed: CUBEUrl
    plugin: CUBEUrl
    descendants: CUBEUrl
    files: CUBEUrl
    parameters: CUBEUrl
    compute_resource: CUBEUrl
    splits: CUBEUrl

    previous_id: Optional[int] = None
    """
    FS plugins will not produce a ``previous_id`` value
    (even though they will return ``"previous": null``)
    """

    size: Optional[int] = None
    """
    IDK what it is the size of.
    
    This field shows up when the plugin instance is maybe done,
    but not when the plugin instance is created.
    """
    template: Optional[dict] = None
    """
    Present only when getting a plugin instance.
    """

    def get_feed(self) -> Feed:
        inst_res = self.s.get(self.url).json()
        feed_url = inst_res['feed']
        feed_res = self.s.get(feed_url).json()
        return Feed(s=self.s, **feed_res)

    def get_start_date(self) -> datetime:
        return datetime.fromisoformat(self.start_date)

    def get_end_date(self) -> datetime:
        return datetime.fromisoformat(self.end_date)



import typer

from caw.commands.store import app, build_client, login_manager
from caw.constants import DEFAULT_ADDRESS


@app.command()
def logout():
    """
    Remove your login credentials.
    """
    addr = None if build_client.address == DEFAULT_ADDRESS else build_client.address
    if login_manager.get(addr) is None:
        typer.echo('Not logged in.', err=True)
    else:
        login_manager.logout(addr)



from typing import NewType, Literal, Union

CUBEToken = NewType('CUBEToken', str)
CUBEAddress = NewType('CUBEAddress', str)
CUBEUsername = NewType('CUBEUsername', str)
CUBEPassword = NewType('CUBEPassword', str)

CUBEUrl = NewType('CUBEUrl', str)

PluginId = NewType('PluginId', int)
PluginName = NewType('PluginName', str)
PluginVersion = NewType('PluginVersion', str)
PluginType = Literal['ds', 'fs', 'ts']

SwiftPath = NewType('SwiftPath', str)

ISOFormatDateString = NewType('ISOFormatDateString', str)

# TODO: PluginInstanceStatus should be an enum
PluginInstanceStatus = Literal[
    'created', 'waiting', 'scheduled', 'started', 'registeringFiles',
    'finishedSuccessfully', 'finishedWithError', 'cancelled'
]

CUBEErrorCode = NewType('CUBEErrorCode', str)

ContainerImageTag = NewType('ContainerImageTag', str)

FeedId = NewType('FeedId', int)
PipingId = NewType('PipingId', int)
PipelineId = NewType('PipelineId', int)

FilesUrl = NewType('FilesUrl', CUBEUrl)

ParameterName = NewType('ParameterName', str)
ParameterType = Union[str, int, float, bool]
ParameterTypeName = Literal['string', 'integer', 'float', 'boolean']
PipelineParameterId = NewType('ParameterLocalId', int)
PluginParameterId = NewType('ParameterGlobalId', int)
PluginInstanceId = NewType('PluginInstanceId', int)

ComputeResourceName = NewType('ComputeResourceName', str)

FileResourceName = NewType('FileResourceName', str)
FileResourceUrl = NewType('FileResourceUrl', str)
FileId = NewType('FileId', int)


from caw.commands.store import app

# register subcommands into app
import caw.commands.base
import caw.commands.login
import caw.commands.logout
import caw.commands.upload
import caw.commands.download
import caw.commands.pipeline
import caw.commands.search



from dataclasses import dataclass, field
from typing import Generator, Collection, Dict, Tuple
from collections import deque
from chris.types import CUBEUrl, ParameterType, PluginInstanceId
from chris.cube.plugin import Plugin
from chris.cube.plugin_instance import PluginInstance
from chris.cube.resource import ConnectedResource


@dataclass(frozen=True)
class PluginTree(ConnectedResource, Collection['PluginTree']):
    """
    A ``PluginTree`` is an immutable node of a directed acyclic graph
    of plugins and default parameters for each plugin.
    It usually represents a runnable pipeline.

    CONSTRAINT: all plugins must be associated with the same CUBE.
    """

    plugin: CUBEUrl
    default_parameters: Dict[str, ParameterType]
    children: Tuple['PluginTree', ...] = field(default_factory=tuple)
    # tuple instead of frozenset because PluginTree.default_parameters
    # is a dict, which is not hashable

    def get_plugin(self) -> Plugin:
        res = self.s.get(self.plugin)
        res.raise_for_status()
        return Plugin(s=self.s, **res.json())

    def run(self, plugin_instance_id: PluginInstanceId
            ) -> Generator[PluginInstance, None, None]:
        """
        Create plugin instances in DFS-order.
        The returned iterator must be iterated through to
        schedule this entire plugin tree.

        :param plugin_instance_id: parent plugin instance
        """
        params = {
            'previous_id': plugin_instance_id
        }
        params.update(self.default_parameters)
        created_instance = self.get_plugin().create_instance(params)
        yield created_instance
        for child in self.children:
            yield from child.run(created_instance.id)

    # the two traversal methods below are not currently used

    def dfs(self) -> Generator['PluginTree', None, None]:
        """
        Depth-first graph traversal.
        """
        yield self
        yield from self.children

    def bfs(self) -> Generator['PluginTree', None, None]:
        """
        Breadth-first graph traversal.

        BFS is insignificantly better than DFS because sibling
        plugin instances can be scheduled insignificantly sooner.
        A stupidly optimal solution would schedule branches by
        doing the HTTP POST requests in parallel.
        """
        queue: deque['PluginTree'] = deque()
        queue.append(self)
        while queue:
            current = queue.popleft()
            yield current
            queue.extend(current.children)

    def __iter__(self):
        return self.dfs()

    def __len__(self):
        count = 0
        for _ in self:
            count += 1
        return count

    def __contains__(self, __x: object) -> bool:
        return any(__x == e for e in self)



"""
An end-to-end test which performs the following:

1. creates a ChRIS user account.
2. caw login
3. caw search
4. caw upload --pipeline ...
5. caw download
6. caw logout
"""

import os
import unittest

import random
import string
import requests
import subprocess as sp
from tempfile import NamedTemporaryFile, TemporaryDirectory
from time import sleep
from glob import iglob


def random_string(length=12) -> str:
    return ''.join(random.choice(string.ascii_letters) for x in range(length))


address = 'http://localhost:8000/api/v1/'
username = 'caw_test_' + random_string(6)
password = random_string(12)


def create_account():
    res = requests.post(
        f'{address}users/',
        headers={
            'Content-Type': 'application/vnd.collection+json',
            'Accept': 'application/json'
        },
        json={
            'template': {
                'data': [
                    {
                        'name': 'email',
                        'value': f'{username}@babyMRI.org'
                    },
                    {
                        'name': 'username',
                        'value': username
                    },
                    {
                        'name': 'password',
                        'value': password
                    }
                ]
            }
        }
    )
    res.raise_for_status()
    data = res.json()

    assert 'username' in data
    assert data['username'] == username


def poll_feed(feed_url: str, jobs_count: int, poll_interval=10, timeout=300) -> dict:
    timer = 0
    headers = {
        'Accept': 'application/json'
    }
    data = {}
    while timer <= timeout:

        print(f'calling get with {feed_url}')
        import logging
        logging.getLogger().setLevel(logging.DEBUG)
        res = requests.get(feed_url, headers=headers, auth=(username, password))
        res.raise_for_status()
        data = res.json()
        if data['finished_jobs'] == jobs_count:
            return data
        sleep(poll_interval)
        timer += poll_interval
    return data


class TestEndToEnd(unittest.TestCase):
    @unittest.skipUnless('CAW_TEST_FULL' in os.environ, 'Set CAW_TEST_FULL=y to run the end-to-end test.')
    def test_endtoend(self):
        create_account()

        sp.run(['caw', '--address', address, '--username', username, 'login', '--password-stdin'],
               input=(password + '\n'), text=True, check=True)

        with NamedTemporaryFile('w', suffix='.txt', delete=False) as f:
            f.write("If you steal from one author it's plagiarism; if you steal from"
                    "\nmany it's research."
                    '\n                -- Wilson Mizner\n')

        search = sp.check_output(['caw', 'search'], text=True)
        self.assertIn('Example branching pipeline', search,
                      msg='"Example branching pipeline" not found in `caw search`')

        feed_url = sp.check_output(['caw', 'upload', '--pipeline', 'Example branching pipeline', '--', f.name],
                                   text=True).rstrip('\n')
        self.assertTrue(feed_url.startswith(address),
                        msg='Feed URL was not correctly printed after `caw upload`.\n'
                            f'output = "{feed_url}", expected to start with "{address}"')
        self.assertTrue(feed_url.endswith('/'),
                        msg='Feed URL was not correctly printed after `caw upload`.\n'
                            f'output = {feed_url}')
        feed_data = poll_feed(feed_url, jobs_count=9)

        with TemporaryDirectory() as tmpdir:
            sp.run(['caw', 'download', feed_data['files'], tmpdir])
            # the pipeline runs pl-dircopy --prefix L where L is a letter.
            # if the DAG is constructed properly, it should produce the following prefixes
            prefixes = {'', 'a', 'ba', 'ca', 'dba', 'eca', 'fca', 'gca', 'hgca'}
            suffix = os.path.basename(f.name)
            results = [
                # example:
                # '/tmp/folder/caw_test_SzvEhj/feed_10/pl-dircopy_81/pl-simpledsapp_82/pl-simpledsapp_83/pl-simpledsapp_85/data/fcatmpl_hy4m5o.txt'
                # --> 'fca'
                os.path.basename(fname[:-len(suffix)])
                for fname in iglob(os.path.join(tmpdir, '**', '*' + suffix), recursive=True)
            ]
            self.assertEqual(len(results), 9, msg='Incorrect number of files produced by feed.')
            self.assertSetEqual(prefixes, set(results),
                                msg='DAG not reconstructed in the correct order.')
        sp.run(['caw', 'logout'])


if __name__ == '__main__':
    unittest.main()



import typer

from caw.commands.store import app, build_client
from caw.run_pipeline import run_pipeline_with_progress


@app.command()
def pipeline(name: str = typer.Argument(..., help='Name of pipeline to run.'),
             target: str = typer.Option(..., help='Plugin instance ID or URL.')):
    """
    Run a pipeline on an existing feed.
    """
    client = build_client()
    plugin_instance = client.get_plugin_instance(target)
    chris_pipeline = client.get_pipeline(name)
    run_pipeline_with_progress(chris_pipeline=chris_pipeline, plugin_instance=plugin_instance)



from pathlib import Path
from typing import Optional, List
import logging

import typer
import requests

from chris.cube.pipeline import Pipeline
from caw.commands.store import app, build_client
from caw.movedata import upload as cube_upload
from caw.run_pipeline import run_pipeline_with_progress

logger = logging.getLogger(__name__)


@app.command()
def upload(
        threads: int = typer.Option(4, '--threads', '-t', help='Number of threads to use for file upload.'),
        create_feed: bool = typer.Option(True, help='Run pl-dircopy on the newly uploaded files.'),
        name: str = typer.Option('', '--name', '-n', help='Name of the feed.'),
        description: str = typer.Option('', '--description', '-d', help='Description of the feed.'),
        pipeline_name: str = typer.Option('', '--pipeline', '-p', help='Name of pipeline to run on the data.'),
        files: List[Path] = typer.Argument(..., help='Files to upload. '
                                                     'Folder upload is supported, but directories are destructured.')
):
    """
    Upload local files and run pl-dircopy.
    """
    client = build_client()
    chris_pipeline: Optional[Pipeline] = None
    if pipeline_name:
        chris_pipeline = client.get_pipeline(pipeline_name)

    try:
        swift_path = cube_upload(client=client, files=files, upload_threads=threads)
    except requests.exceptions.RequestException as e:
        logger.debug('RequestException: %s\n%s', str(e), e.response.text)
        typer.secho('Upload unsuccessful', fg=typer.colors.RED, err=True)
        raise typer.Abort()

    if not create_feed:
        raise typer.Exit()

    dircopy_instance = client.run('pl-dircopy', params={'dir': swift_path})
    if name:
        dircopy_instance.get_feed().set_name(name)
    if description:
        dircopy_instance.get_feed().set_description(description)

    if chris_pipeline:
        run_pipeline_with_progress(chris_pipeline=chris_pipeline, plugin_instance=dircopy_instance)
    typer.echo(dircopy_instance.feed)



from pathlib import Path
from dataclasses import dataclass, field
import requests
from typing import Optional, Union, Generator, Dict

from chris.types import (
    CUBEAddress, CUBEToken, CUBEUsername, CUBEPassword, CUBEUrl, PluginInstanceId, PluginName, PluginVersion
)
from chris.cube.plugin import Plugin
from chris.cube.plugin_instance import PluginInstance
from chris.cube.files import ListOfDownloadableFiles
from chris.cube.registered_pipeline import RegisteredPipeline
from chris.cube.pagination import fetch_paginated_objects
from chris.cube.resource import ConnectedResource
from chris.errors import (
    ChrisClientError, ChrisIncorrectLoginError, PluginNotFoundError, PipelineNotFoundError
)
from chris.helpers.peek import peek

import logging

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ChrisClient(ConnectedResource):

    address: CUBEAddress
    token: CUBEToken

    collection_links: Dict[str, CUBEUrl] = field(init=False)
    s: requests.Session = field(init=False)
    url: CUBEUrl = field(init=False)
    """
    An alias for ``address``
    """

    def __post_init__(self):
        if not self.address.endswith('/api/v1/'):
            raise ValueError('Address of CUBE must end with "/api/v1/"')
        object.__setattr__(self, 'url', self.address)
        object.__setattr__(self, 's', self.__start_session(self.token))
        object.__setattr__(self, 'collection_links', self.__get_collection_links())

    @classmethod
    def from_login(cls,
                   address: Union[CUBEAddress, str],
                   username: Union[CUBEUsername, str],
                   password: Union[CUBEPassword, str]) -> 'ChrisClient':
        login = requests.post(address + 'auth-token/', json={
            'username': username,
            'password': password
        })
        if login.status_code == 400:
            res = login.json()
            raise ChrisIncorrectLoginError(
                res['non_field_errors'][0] if 'non_field_errors' in res else login.text
            )
        login.raise_for_status()
        return cls(address=address, token=login.json()['token'])

    @staticmethod
    def __start_session(token) -> requests.Session:
        s = requests.Session()
        s.headers.update({
            'Accept': 'application/json',
            'Content-Type': 'application/vnd.collection+json',
            'Authorization': 'Token ' + token
        })
        return s

    def __get_collection_links(self) -> Dict[str, CUBEUrl]:
        """
        Make a request to the CUBE address. Calling this method verifies
        that the login token is correct.
        """
        res = self.s.get(self.address)
        if res.status_code == 401:
            data = res.json()
            raise ChrisIncorrectLoginError(data['detail'] if 'detail' in data else res.text)
        if res.status_code != 200:
            raise ChrisClientError(f'CUBE response status code was {res.status_code}.')
        res.raise_for_status()
        data = res.json()
        if 'collection_links' not in data or 'uploadedfiles' not in data['collection_links']:
            raise ChrisClientError(f'Unexpected CUBE response: {res.text}')
        return data['collection_links']

    @property
    def search_addr_plugins(self) -> CUBEUrl:
        return CUBEUrl(self.address + 'plugins/search/')

    @property
    def search_addr_plugins_instances(self) -> CUBEUrl:
        return CUBEUrl(self.address + 'plugins/instances/search/')

    @property
    def search_addr_pipelines(self) -> CUBEUrl:
        return CUBEUrl(self.address + 'pipelines/search/')

    def upload(self, file_path: Path, upload_folder: Path) -> dict:
        """
        Upload a local file into ChRIS backend Swift storage.

        :param file_path: local file path
        :param upload_folder: path in Swift where to upload to
        :return: response
        """
        upload_path = upload_folder / file_path.name

        with open(file_path, 'rb') as file_object:
            files = {
                'upload_path': (None, str(upload_path)),
                'fname': (file_path.name, file_object)
            }
            res = self.s.post(
                self.collection_links['uploadedfiles'],
                files=files,
                headers={
                    'Accept': 'application/vnd.collection+json',
                    'Content-Type': None
                }
            )
        res.raise_for_status()
        return res.json()

    def get_plugin(self, name_exact='', version='', url='') -> Plugin:
        """
        Get a single plugin, either searching for it by its exact name, or by URL.

        :param name_exact: name of plugin
        :param version: (optional) version of plugin
        :param url: (alternative to name_exact) url of plugin
        """
        if url:
            return self.get_plugin_by_url(url)
        return self.get_plugin_by_name(name_exact, version)

    def get_plugin_by_url(self, url: Union[CUBEUrl, str]):
        res = self.s.get(url)
        res.raise_for_status()
        return Plugin(**res.json(), s=self.s)

    def get_plugin_by_name(self, name_exact: Union[PluginName, str],
                           version: Optional[Union[PluginVersion, str]] = None):
        search = self.search_plugin(name_exact, version)
        return peek(search, mt=PluginNotFoundError)

    def search_plugin(self, name_exact: str, version: Optional[str] = None
                      ) -> Generator[Plugin, None, None]:
        qs = self._join_qs(name_exact=name_exact, version=version)
        url = CUBEUrl(f'{self.search_addr_plugins}?{qs}')
        return fetch_paginated_objects(s=self.s, url=url, constructor=Plugin)

    def get_plugin_instance(self, plugin: Union[CUBEUrl, PluginInstanceId]):
        """
        Get a plugin instance.
        :param plugin: Either a plugin instance ID or URL
        :return: plugin instance
        """
        url = plugin if '/' in plugin else f'{self.address}plugins/instances/{plugin}/'
        res = self.s.get(url)
        res.raise_for_status()
        return PluginInstance(**res.json(), s=self.s)

    def run(self, plugin_name='', plugin_url='', plugin: Optional[PluginInstance] = None,
            params: Optional[dict] = None) -> PluginInstance:
        """
        Create a plugin instance. Either provide a plugin object,
        or search for a plugin by name or URL.
        :param plugin: plugin to run
        :param plugin_name: name of plugin to run
        :param plugin_url: alternatively specify plugin URL
        :param params: plugin parameters as key-value pairs (not collection+json)
        :return:
        """
        if not plugin:
            plugin = self.get_plugin(name_exact=plugin_name, url=plugin_url)
        return plugin.create_instance(params)

    def search_uploadedfiles(self, fname='', fname_exact='') -> ListOfDownloadableFiles:
        qs = self._join_qs(fname=fname, fname_exact=fname_exact)
        url = CUBEUrl(f"{self.collection_links['uploadedfiles']}search/?{qs}")
        return self.get_files(url)

    def get_files(self, url: CUBEUrl) -> ListOfDownloadableFiles:
        return ListOfDownloadableFiles(url=url, s=self.s)

    def search_pipelines(self, name='') -> Generator[RegisteredPipeline, None, None]:
        return fetch_paginated_objects(
            s=self.s,
            url=CUBEUrl(f"{self.collection_links['pipelines']}search/?name={name}"),
            constructor=RegisteredPipeline
        )

    def get_pipeline(self, name: str) -> RegisteredPipeline:
        return peek(self.search_pipelines(name), mt=PipelineNotFoundError)

    def get_username(self) -> CUBEUsername:
        res = self.s.get(url=self.collection_links['user'])
        res.raise_for_status()
        data = res.json()
        return CUBEUsername(data['username'])

    @staticmethod
    def _join_qs(**kwargs) -> str:
        return '&'.join([f'{k}={v}' for k, v in kwargs.items() if v])



"""
A global variable namespace.
"""

import typer
from caw.login.manager import LoginManager
from caw.builder import ClientBuilder


login_manager = LoginManager()
build_client = ClientBuilder(login_manager)

app = typer.Typer(
    epilog='Examples and documentation at '
           'https://github.com/FNNDSC/caw#documentation'
)



from pytest_mock import MockerFixture
from chris.cube.registered_pipeline import Piping, MutablePluginTreeNode
from chris.cube.plugin_tree import PluginTree
from chris.types import PluginId, PipingId, PipelineId, CUBEUrl, ParameterName


r"""
    a --bow wow
   / \
  b   c
     / \
    d   e --this isit
"""


def test_freeze(mocker: MockerFixture):
    session = mocker.Mock()
    ep = Piping(
        s=session,
        id=PipingId(5),
        url=CUBEUrl('https://example.com/api/v1/pipelines/3/pipings/5/'),
        previous=PipingId(2),
        plugin_id=PluginId(20),
        plugin=CUBEUrl('https://example.com/api/v1/plugins/20/'),
        pipeline_id=PipelineId(3),
        pipeline=CUBEUrl('https://example.com/api/v1/pipelines/3/')
    )
    dp = Piping(
        s=session,
        id=PipingId(4),
        url=CUBEUrl('https://example.com/api/v1/pipelines/3/pipings/4/'),
        previous=PipingId(2),
        plugin_id=PluginId(30),
        plugin=CUBEUrl('https://example.com/api/v1/plugins/30/'),
        pipeline_id=PipelineId(3),
        pipeline=CUBEUrl('https://example.com/api/v1/pipelines/3/')
    )
    cp = Piping(
        s=session,
        id=PipingId(2),
        url=CUBEUrl('https://example.com/api/v1/pipelines/3/pipings/2/'),
        previous=PipingId(1),
        plugin_id=PluginId(40),
        plugin=CUBEUrl('https://example.com/api/v1/plugins/40/'),
        pipeline_id=PipelineId(3),
        pipeline=CUBEUrl('https://example.com/api/v1/pipelines/3/')
    )
    bp = Piping(
        s=session,
        id=PipingId(3),
        url=CUBEUrl('https://example.com/api/v1/pipelines/3/pipings/3/'),
        previous=PipingId(1),
        plugin_id=PluginId(50),
        plugin=CUBEUrl('https://example.com/api/v1/plugins/50/'),
        pipeline_id=PipelineId(3),
        pipeline=CUBEUrl('https://example.com/api/v1/pipelines/3/')
    )
    ap = Piping(
        s=session,
        id=PipingId(1),
        url=CUBEUrl('https://example.com/api/v1/pipelines/3/pipings/1/'),
        previous=None,
        plugin_id=PluginId(60),
        plugin=CUBEUrl('https://example.com/api/v1/plugins/60/'),
        pipeline_id=PipelineId(3),
        pipeline=CUBEUrl('https://example.com/api/v1/pipelines/3/')
    )

    em = MutablePluginTreeNode(s=session, piping=ep, params={ParameterName('this'): 'isit'})
    dm = MutablePluginTreeNode(s=session, piping=dp, params={})
    cm = MutablePluginTreeNode(s=session, piping=cp, params={})
    bm = MutablePluginTreeNode(s=session, piping=bp, params={})
    am = MutablePluginTreeNode(s=session, piping=ap, params={ParameterName('bow'): 'wow'})

    cm.children.append(dm)
    cm.children.append(em)
    am.children.append(bm)
    am.children.append(cm)

    et = PluginTree(s=session, plugin=ep.plugin, default_parameters={ParameterName('this'): 'isit'})
    dt = PluginTree(s=session, plugin=dp.plugin, default_parameters={})
    ct = PluginTree(s=session, plugin=cp.plugin, default_parameters={},
                    children=(dt, et))
    bt = PluginTree(s=session, plugin=bp.plugin, default_parameters={})
    at = PluginTree(s=session, plugin=ap.plugin, default_parameters={ParameterName('bow'): 'wow'},
                    children=(bt, ct))

    assert at == am.freeze()



import abc
from dataclasses import dataclass
from chris.types import CUBEUrl
from chris.cube.resource.connected_resource import ConnectedResource


@dataclass(frozen=True)
class CUBEResource(ConnectedResource, abc.ABC):
    """
    A ``CUBEResource`` represents data returned from an endpoint of the CUBE API.

    Responses from the CUBE API always return a ``url`` which indicated the requested URI.
    """
    url: CUBEUrl



import unittest
from unittest.mock import Mock, patch

import typer

from chris.types import CUBEAddress, CUBEToken, CUBEUsername, CUBEPassword
from chris.errors import ChrisIncorrectLoginError
from caw.builder import ClientBuilder
from caw.login.manager import LoginManager


class ClientBuilderTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.lm = Mock(spec=LoginManager)
        self.builder = ClientBuilder(self.lm)

    # @patch('caw.builder.FriendlyClient')
    # def test_using_default_session(self, constructor):
    #     address = CUBEAddress('https://example.com/api/v1/')
    #     token = CUBEToken('my_token_value')
    #     self.lm.get_default_address.return_value = address
    #     self.lm.get.return_value = token
    #
    #     self.builder()
    #     constructor.assert_called_once_with(address, token=token)

    # @patch('caw.builder.FriendlyClient')
    # def test_using_token_and_specific_address(self, constructor):
    #     address = CUBEAddress('https://example.com/api/v1/')
    #     token = CUBEToken('my_token_value')
    #     self.lm.get.return_value = token
    #     self.builder.address = address
    #
    #     self.builder()
    #     constructor.assert_called_once_with(address, token=token)
    #     self.lm.get_default_address.assert_not_called()

    # @patch('caw.builder.FriendlyClient')
    # def test_using_username_password(self, constructor):
    #     self.builder.address = CUBEAddress('https://example.com/api/v1/')
    #     self.builder.username = CUBEUsername('crab')
    #     self.builder.password = CUBEPassword('shrimp')
    #
    #     self.builder()
    #     constructor.assert_called_once_with(self.builder.address,
    #                                         username=self.builder.username,
    #                                         password=self.builder.password)
    #     self.lm.get_default_address.assert_not_called()
    #     self.lm.get.assert_not_called()

    @patch('typer.secho', spec=typer.secho)
    @patch('chris.client.ChrisClient.__init__', side_effect=ChrisIncorrectLoginError)
    def test_logout_after_token_expired(self, _, secho):
        address = CUBEAddress('https://example.com/api/v1/')
        self.lm.get_default_address.return_value = address
        self.lm.get.return_value = CUBEToken('my_token_value')

        with self.assertRaises(typer.Abort):
            self.builder()

        self.lm.get_default_address.assert_called_once()
        self.lm.get.assert_called_once_with(address)
        self.lm.logout.assert_called_once_with(address)
        secho.assert_called_with(
            f'warning: removing saved login for {address}.\n'
            f'To login again, run\n'
            f"\n\tcaw --address '{address}' login",
            err=True
        )


class UserErrorsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.lm = Mock(spec=LoginManager)
        self.builder = ClientBuilder(self.lm)

    @patch('typer.echo', spec=typer.echo)
    def test_given_username_no_password(self, echo):
        self.builder.username = CUBEUsername('crab')
        with self.assertRaises(typer.Abort):
            self.builder()
        echo.assert_called_with('Given a username but no password. You must supply both.', err=True)

    @patch('typer.echo', spec=typer.echo)
    def test_given_password_no_username(self, echo):
        self.builder.password = CUBEPassword('crab')
        with self.assertRaises(typer.Abort):
            self.builder()
        echo.assert_called_with('Given a password but no username. You must supply both.', err=True)


if __name__ == '__main__':
    unittest.main()



from chris.types import CUBEAddress, CUBEUsername, CUBEPassword

DEFAULT_ADDRESS = CUBEAddress('http://localhost:8000/api/v1/')
DEFAULT_USERNAME = CUBEUsername('chris')
DEFAULT_PASSWORD = CUBEPassword('chris1234')



import sys
import typer
from chris.types import CUBEUsername, CUBEPassword
from caw.commands.store import app, build_client, login_manager


@app.command()
def login(read_pass: bool = typer.Option(False, '--password-stdin', help='Take the password from stdin')):
    """
    Login to ChRIS.
    """

    if not build_client.username:
        build_client.username = CUBEUsername(typer.prompt('username'))

    if not build_client.password:
        if read_pass:
            build_client.password = CUBEPassword(sys.stdin.read().rstrip('\n'))
        else:
            build_client.password = typer.prompt('password', hide_input=True)

    client = build_client()
    login_manager.login(client.address, client.token)



from dataclasses import dataclass
from chris.cube.resource import CUBEResource
from chris.types import (
    PluginId, PluginName, PluginType, PluginVersion, ISOFormatDateString, ContainerImageTag, CUBEUrl
)
from chris.cube.plugin_instance import PluginInstance
from chris.helpers.collection import collection_helper

import logging

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Plugin(CUBEResource):
    """
    A *plugin* in *ChRIS* describes a unit of compute.
    To run something on *ChRIS*, a user creates a :class:`PluginInstance`
    of a *plugin*.
    """
    id: PluginId
    creation_date: ISOFormatDateString
    name: PluginName

    version: PluginVersion
    dock_image: ContainerImageTag
    public_repo: str
    icon: str
    type: PluginType
    stars: int

    authors: str
    title: str
    category: str
    description: str
    documentation: str
    license: str

    execshell: str
    selfpath: str
    selfexec: str
    min_number_of_workers: int
    max_number_of_workers: int
    min_cpu_limit: int
    max_cpu_limit: int
    min_gpu_limit: int
    max_gpu_limit: int
    min_memory_limit: int
    max_memory_limit: int

    meta: CUBEUrl
    parameters: CUBEUrl
    instances: CUBEUrl
    compute_resources: CUBEUrl

    def create_instance(self, params: dict = None) -> PluginInstance:
        logging.debug('%s: %s', self.name, params)
        payload = collection_helper(params)
        res = self.s.post(self.instances, json=payload)
        res.raise_for_status()
        return PluginInstance(**res.json(), s=self.s)



import os
import sys
from concurrent.futures import ThreadPoolExecutor
import datetime
from typing import List, Tuple, Union
import typer
from chris.client import ChrisClient
from chris.cube.files import DownloadableFile
from chris.types import CUBEUrl
import logging
from pathlib import Path


def upload(client: ChrisClient, files: List[Path], parent_folder='', upload_threads=4):

    username = client.get_username()

    if parent_folder:
        upload_folder = f'{username}/uploads/{parent_folder}/{datetime.datetime.now().isoformat()}/'
    else:
        upload_folder = f'{username}/uploads/{datetime.datetime.now().isoformat()}/'

    input_files: List[Path] = []
    for path in files:
        if path.is_file():
            input_files.append(path)
        elif path.is_dir():
            nested_files = [f for f in path.rglob('**/*') if f.is_file()]
            if len(nested_files) > 0:
                input_files.extend(nested_files)
            else:
                typer.secho(f'WARNING: input directory is empty: {path}', dim=True, err=True)
        else:
            typer.secho(f'No such file or directory: {path}', fg=typer.colors.RED, err=True)
            raise typer.Abort()

    if len(input_files) == 0:
        typer.secho(f'No input files specified.', fg=typer.colors.RED, err=True)
        raise typer.Abort()

    with typer.progressbar(label='Uploading files', length=len(input_files), file=sys.stderr) as bar:
        def upload_file(input_file: str):
            client.upload(Path(input_file), Path(upload_folder))
            bar.update(1)

        with ThreadPoolExecutor(max_workers=upload_threads) as pool:
            uploads = pool.map(upload_file, input_files)

    # check for upload errors
    for upload_result in uploads:
        logging.debug(upload_result)

    typer.secho(f'Successfully uploaded {len(input_files)} files to "{upload_folder}"', fg=typer.colors.GREEN, err=True)
    return upload_folder


def download(client: ChrisClient, url: Union[str, CUBEUrl], destination: Path, threads: 4):
    """
    Download all the files from a given ChRIS API url.
    :param client: ChRIS client
    :param url: any ChRIS file resource url, e.g.
                https://cube.chrisproject.org/api/v1/uploadedfiles/
                https://cube.chrisproject.org/api/v1/uploadedfiles/?fname=chris/uploads/today
                https://cube.chrisproject.org/api/v1/3/files/
    :param destination: folder on host where to download to
    :param threads: max number of concurrent downloads
    """
    if destination.is_file():
        typer.secho(f'Cannot download into {destination}: is a file', fg=typer.colors.RED, err=True)
        raise typer.Abort()

    def __calculate_target(remote_file: DownloadableFile) -> Tuple[Path, DownloadableFile]:
        """
        Decide on a download location for a file resource in ChRIS.
        Create the parent directory if needed.
        :param remote_file: file information from ChRIS
        :return: download location on host and that file
        """
        fname = remote_file.fname
        if fname.startswith('chris/'):
            fname = fname[6:]
        target = destination.joinpath(fname)
        os.makedirs(target.parent, exist_ok=True)
        return target, remote_file

    search = tuple(client.get_files(url))
    with typer.progressbar(search, length=len(search), label='Getting information', file=sys.stderr) as progress:
        to_download = frozenset(__calculate_target(remote_file) for remote_file in progress)

    with typer.progressbar(length=len(to_download), label='Downloading files', file=sys.stderr) as progress:
        def download_file(t: Tuple[Path, DownloadableFile]) -> int:
            """
            Download file and move the progress bar
            :param t: tuple
            :return: downloaded file size
            """
            target, remote_file = t
            remote_file.download(target)
            progress.update(1)
            return target.stat().st_size

        with ThreadPoolExecutor(max_workers=threads) as pool:
            sizes = pool.map(download_file, to_download)

    total_size = sum(sizes)
    if total_size < 2e5:
        size = f'{total_size} bytes'
    elif total_size < 2e8:
        size = f'{total_size / 1e6:.4f} MB'
    else:
        size = f'{total_size / 1e9:.4f} GB'
    typer.secho(size, fg=typer.colors.GREEN, err=True)



import time
from typing import Callable, TypeVar, Iterable, Optional, Union, FrozenSet, Generator, Type
import requests

from chris.types import CUBEAddress
from chris.errors import WaitTimeoutException


T = TypeVar('T')


def __optional2set(c: Optional[Union[T, Iterable[T]]]) -> FrozenSet[T]:
    if c is None:
        c = []
    elif not isinstance(c, Iterable):
        c = [c]
    return frozenset(c)


def block_until(poller: Callable[[], T],
                not_ready_exception: Optional[Union[Type[Exception], Iterable[Type[Exception]]]] = None,
                interval: float = 2.0,
                timeout: float = 300
                ) -> Generator[Union[Exception, T], None, None]:
    """
    Blocks until a "ready" event by repeatedly polling using a function.

    The given polling function may either raise an allowed exception or produce a return value.

    A client should iterate over ``block_until`` using a loop, and break when ``block_until``
    yields an acceptable return value.

    Example:

    .. code-block:: python

        for poll in block_until(poller=do_request):
            if poll == 'OK':
                break


    :param poller: the function to call which polls the service
    :param not_ready_exception: a set of Exceptions which the poller may raise, indicating the
                                service is not "ready"
    :param interval: number of seconds to wait between poll
    :param timeout: maximum amount of time to wait
    :return: a Generator which yields the exceptions or return values produced when invoking the poller
    """

    if timeout <= 0:
        raise ValueError('timeout must be in the range (0, inf)')
    if interval <= 0:
        raise ValueError('interval must be in the range (0, inf)')

    permissible_exceptions = __optional2set(not_ready_exception)

    elapsed = 0
    while elapsed <= timeout:
        try:
            yield poller()
        except Exception as e:
            if e.__class__ in permissible_exceptions:
                yield e
                continue
            raise e
        finally:
            time.sleep(interval)
            elapsed += interval
    raise WaitTimeoutException(f'Timeout reached after {timeout} seconds')


def __expected_response_from_users(users_url: str) -> dict:
    return {
        'collection': {
            'version': '1.0',
            'href': users_url,
            'items': [],
            'links': [],
            'template': {
                'data': [
                    {'name': 'username', 'value': ''},
                    {'name': 'password', 'value': ''},
                    {'name': 'email', 'value': ''}
                ]
            },
            'total': 0
        }
    }


def wait_until_ready(url: CUBEAddress) -> Generator[None, None, None]:
    """
    Wait for the ChRIS backend service to be ready to accept connections by polling the
    ``/api/v1/users/`` endpoint.

    :param url: address of ChRIS backend
    :return: a Generator which produces the errors or responses from polling
    """
    if url.endswith('api/v1/'):
        url += 'users/'

    session = requests.Session()
    session.headers.update({
        'Accept': 'application/vnd.collection+json'
    })

    expected = __expected_response_from_users(url)

    def poll_user() -> bool:
        return session.get(url=url).json()

    for res in block_until(poller=poll_user, not_ready_exception=requests.exceptions.ConnectionError,
                           interval=2.0, timeout=300.0):
        if isinstance(res, dict) and res == expected:
            break
        yield res



from pathlib import Path
from dataclasses import dataclass
from typing import Iterable, Generator
import requests

from chris.cube.resource import CUBEResource
from chris.cube.pagination import fetch_paginated_objects
from chris.types import FileResourceUrl, FileResourceName


# TODO add all the fields of the various kinds of files
# UploadedFile, FeedFile, PACSFile


@dataclass(frozen=True)
class DownloadableFile(CUBEResource):
    file_resource: FileResourceUrl
    fname: FileResourceName

    def download(self, destination: Path, chunk_size=8192):
        with self.s.get(self.file_resource, stream=True, headers={'Accept': None}) as r:
            r.raise_for_status()
            with destination.open('wb') as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    f.write(chunk)


@dataclass(frozen=True)
class ListOfDownloadableFiles(Iterable[DownloadableFile], CUBEResource):
    def __iter__(self) -> Generator[DownloadableFile, None, None]:
        return fetch_paginated_objects(s=self.s, url=self.url, constructor=self._construct_downloadable_file)

    @staticmethod
    def _construct_downloadable_file(s: requests.Session, **kwargs):
        return DownloadableFile(s=s, url=kwargs['url'],
                                fname=kwargs['fname'],
                                file_resource=kwargs['file_resource'])



class ChrisClientError(Exception):
    pass


class ChrisIncorrectLoginError(ChrisClientError):
    pass


class ChrisResourceNotFoundError(ChrisClientError):
    pass


class PluginNotFoundError(ChrisResourceNotFoundError):
    pass


class PipelineNotFoundError(ChrisResourceNotFoundError):
    pass


class WaitTimeoutException(Exception):
    pass



import pytest
from pytest_mock import MockerFixture
from unittest.mock import Mock

from chris.cube.feed import Feed
from chris.tests.mocks.data.feed import data


@pytest.fixture
def session(mocker: MockerFixture) -> Mock:
    return mocker.Mock()


@pytest.fixture
def feed(session) -> Feed:
    return Feed(s=session, **data)


def test_set_name(session: Mock, feed: Feed):
    feed.set_name('A New Name for my Feed')
    session.put.assert_called_once_with(
        'https://example.com/api/v1/3/',
        json={
            'template': {
                'data': [
                    {
                        'name': 'name',
                        'value': 'A New Name for my Feed'
                    }
                ]
            }
        }
    )


def test_set_description(session: Mock, feed: Feed):
    feed.set_description('A new description for my feed.')
    session.put.assert_called_once_with(
        'https://example.com/api/v1/note3/',
        json={
            'template': {
                'data': [
                    {
                        'name': 'title',
                        'value': 'Description'
                    },
                    {
                        'name': 'content',
                        'value': 'A new description for my feed.'
                    }
                ]
            }
        }
    )



"""
Session token storage mechanisms.

If available, the desktop keyring is used. Else a session token
(not the account password) will be stored in plaintext.
"""
import json
from typing import Optional, Type
import typer
from pathlib import Path
from packaging import version

import caw
from chris.types import CUBEAddress, CUBEToken
from caw.login.store import AbstractSecretStore, KeyringSecretStore, PlaintextSecretStore, use_keyring

PreferredSecretStore: Type[AbstractSecretStore] = KeyringSecretStore if use_keyring else PlaintextSecretStore
"""
``PreferredSecretStore`` is the :class:`KeyringSecretStore` if the :mod:`keyring` module is
available. Otherwise, :class:`PlaintextSecretStore` is used instead.
"""

default_config_file: Path = Path(typer.get_app_dir('caw')) / 'login.json'


class LoginManager:
    """
    A wrapper around an :class:`AbstractSecretStore` which handles saving of configurations.

    The configuration file is saved in ``~/.config/caw`` after each method call.
    """

    VERSION = caw.__version__

    def __init__(self, SecretStoreBackend: Type[AbstractSecretStore] = PreferredSecretStore,
                 config_file: Optional[Path] = default_config_file):
        self.__savefile = config_file
        self._initialize_config()

        with self.__savefile.open('r') as f:
            self._config = json.load(f)

        if 'secretStore' not in self._config:
            self._config['secretStore'] = dict()
        self._store = SecretStoreBackend(self._config['secretStore'])

    def _initialize_config(self):
        """
        Creates an empty configuration file on first run.

        If the configuration file already exists but its spec was written by
        a backwards-incompatible version, the configuration is deleted and
        recreated.
        """
        if self.__savefile.exists():
            if not self._config_file_is_sane():
                typer.secho(f'WARNING: config file {self.__savefile} was created by '
                            f'an incompatible version of {caw.pkg.metadata["name"]}.\n'
                            'It will be overwritten with an empty configuration.',
                            dim=True, err=True)
                self._write_empty_config()
        else:
            self.__savefile.parent.mkdir(parents=True, exist_ok=True)
            self._write_empty_config()

    def _config_file_is_sane(self) -> bool:
        """
        :return: True if the configuration file is valid JSON and its major version
                 spec is compatible with the current version of caw
        """
        with self.__savefile.open('r') as f:
            try:
                data = json.load(f)
            except json.JSONDecodeError:
                msg = f'CRITICAL: the file {self.__savefile} is not valid JSON.'
                typer.secho(msg, fg=typer.colors.RED, err=True)
                raise typer.Abort()

        if 'version' not in data:
            return False

        try:
            spec = version.Version(data['version'])
        except version.InvalidVersion:
            typer.secho(f'WARNING: invalid config version specified in {self.__savefile}.',
                        fg=typer.colors.RED, err=True)
            return False

        return spec.major == self.VERSION.major

    def _write_empty_config(self):
        """
        Overwrites the configuration file with an empty configuration.
        """
        self.__savefile.write_text(f'{{"version": "{self.VERSION}"}}')

    def get_default_address(self, address: Optional[CUBEAddress] = None) -> Optional[CUBEAddress]:
        if address is not None:
            return address
        if 'defaultAddress' not in self._config:
            return None
        return CUBEAddress(self._config['defaultAddress'])

    def _write_config(self):
        """
        Save login configuration to disk.
        """
        with self.__savefile.open('w') as f:
            json.dump(self._config, f)

    def get(self, address: Optional[CUBEAddress] = None) -> Optional[CUBEToken]:
        address = self.get_default_address(address)
        if not address:
            return None
        return CUBEToken(self._store.get(address))

    def logout(self, address: Optional[CUBEAddress] = None):
        """
        Remove secret from storage. If the address is the default address, then remove the default address as well.
        """
        address = self.get_default_address(address)
        if address == self.get_default_address() and 'defaultAddress' in self._config:
            del self._config['defaultAddress']

        self._store.clear(address)
        self._write_config()

    def login(self, address: CUBEAddress, token: CUBEToken):
        self._store.set(address, token)
        self._config['defaultAddress'] = address
        self._write_config()

        if not use_keyring and isinstance(self._store, PlaintextSecretStore):
            typer.secho(f'Login token was saved as plaintext in the file {self.__savefile}',
                        dim=True, err=True)
            typer.secho('For safer credentials storage, please run: '
                        '\n\n\tcaw logout'
                        '\n\tpip install keyring'
                        f'\n\tcaw --address {address} login\n', dim=True, err=True)



from dataclasses import dataclass, field
from typing import Generator, Sequence, Optional, List, Dict
import requests
from chris.cube.resource import CUBEResource
from chris.cube.pagination import fetch_paginated_objects
from chris.cube.pipeline import Pipeline
from chris.cube.plugin_tree import PluginTree

from chris.types import (
    CUBEUrl, PipelineId, CUBEUsername, ISOFormatDateString,
    ParameterName, ParameterType, ParameterTypeName, PipingId,
    PipelineParameterId, PluginParameterId,
    PluginName, PluginVersion, PluginId
)


class PipelineAssemblyException(Exception):
    """
    Pipeline JSON representation cannot be reassembled as a Piping DAG.
    """
    pass


class PipelineHasMultipleRootsException(PipelineAssemblyException):
    """
    Multiple *pipings* with 'previous': null were found in the pipeline JSON representation.
    """
    pass


class PipelineRootNotFoundException(PipelineAssemblyException):
    """
    No piping found in the pipelines JSON representation with 'previous': null.
    """
    pass


@dataclass(frozen=True)
class PipingParameter(CUBEResource):
    id: PipelineParameterId
    value: ParameterType
    type: ParameterTypeName
    plugin_piping_id: PipingId
    previous_plugin_piping_id: None
    param_name: ParameterName
    param_id: PluginParameterId
    plugin_piping: CUBEUrl
    plugin_name: PluginName
    plugin_version: PluginVersion
    plugin_id: PluginId
    plugin_param: CUBEUrl


@dataclass(frozen=True)
class Piping(CUBEResource):
    id: PipingId
    plugin_id: PluginId
    pipeline_id: PipelineId
    previous: CUBEUrl
    plugin: CUBEUrl
    pipeline: CUBEUrl
    previous_id: Optional[PipingId] = None


@dataclass
class MutablePluginTreeNode:
    """
    A mutable predecessor to :class:`PluginTree`.
    """
    s: requests.Session
    piping: Piping
    params: Dict[ParameterName, ParameterType] = field(default_factory=dict)
    children: List['MutablePluginTreeNode'] = field(default_factory=list)

    def freeze(self) -> PluginTree:
        """
        Convert to an immutable :class:`PluginTree`.
        """
        return PluginTree(
            s=self.s,
            plugin=self.piping.plugin,
            default_parameters=self.params,
            children=tuple(n.freeze() for n in self.children)
        )


@dataclass(frozen=True)
class RegisteredPipeline(CUBEResource, Pipeline):
    id: PipelineId
    locked: bool
    owner_username: CUBEUsername
    creation_date: ISOFormatDateString
    modification_date: ISOFormatDateString
    plugins: CUBEUrl
    plugin_pipings: CUBEUrl
    default_parameters: CUBEUrl
    instances: CUBEUrl

    def get_default_parameters(self) -> Sequence[PipingParameter]:
        return list(fetch_paginated_objects(s=self.s, url=self.default_parameters, constructor=PipingParameter))

    @staticmethod
    def map_parameters(params: Sequence[PipingParameter]) -> Dict[PipingId, Dict[ParameterName, ParameterType]]:
        assembled_params: Dict[PipingId, Dict[ParameterName, ParameterType]] = {
            p.plugin_piping_id: {} for p in params
        }
        for p in params:
            assembled_params[p.plugin_piping_id][ParameterName(p.param_name)] = p.value
        return assembled_params

    def get_pipings(self) -> Generator[Piping, None, None]:
        yield from fetch_paginated_objects(s=self.s, url=self.plugin_pipings, constructor=Piping)

    def get_root(self) -> PluginTree:
        """
        Assemble a tree of plugins by making requests to CUBE.
        """
        # collect all default parameters
        assembled_params = self.map_parameters(self.get_default_parameters())
        pipings_map: Dict[PipingId, MutablePluginTreeNode] = {}
        root: Optional[MutablePluginTreeNode] = None

        # create DAG nodes
        for piping in self.get_pipings():
            params = assembled_params[piping.id] if piping.id in assembled_params else {}

            node = MutablePluginTreeNode(piping=piping, params=params, s=self.s)
            pipings_map[piping.id] = node

            if piping.previous:
                pipings_map[piping.previous_id].children.append(node)
            else:
                if root is not None:
                    raise PipelineHasMultipleRootsException()
                root = node
        if not root:
            raise PipelineRootNotFoundException()
        return root.freeze()



import abc
from dataclasses import dataclass
from chris.cube.plugin_tree import PluginTree


@dataclass(frozen=True)
class Pipeline(abc.ABC):
    name: str
    authors: str
    description: str
    category: str

    @abc.abstractmethod
    def get_root(self) -> PluginTree:
        ...



import unittest
from unittest.mock import Mock

import json
from pathlib import Path

from chris.types import CUBEAddress, CUBEToken
from caw.login.store import AbstractSecretStore, KeyringSecretStore, PlaintextSecretStore, use_keyring
from caw.login.manager import LoginManager

from tempfile import NamedTemporaryFile


class TestSecretStore(unittest.TestCase):
    def can_save_clear(self, store: AbstractSecretStore):
        store.set('http://localhost:8910/api/v1/', 'abcdefg')
        stored = store.get('http://localhost:8910/api/v1/')
        self.assertEqual(stored, 'abcdefg',
                         msg='Stored secret does not match what was originally set.')
        store.clear('http://localhost:8910/api/v1/')
        self.assertIsNone(store.get('http://localhost:8910/api/v1/'),
                          msg='store.clear did not work.')

    @unittest.skipUnless(use_keyring, 'keyring not supported')
    def test_keyring(self):
        self.can_save_clear(KeyringSecretStore({}))

    def test_plaintext(self):
        self.can_save_clear(PlaintextSecretStore({}))


class TestLoginManager(unittest.TestCase):

    def setUp(self) -> None:
        with NamedTemporaryFile(suffix='.json', delete=True) as self.savefile:
            pass
        self.savefile = Path(self.savefile.name)
        self.store: Mock = Mock(spec=AbstractSecretStore)
        wrapper = Mock(return_value=self.store, spec=AbstractSecretStore.__init__)
        self.lm = LoginManager(wrapper, self.savefile)  # type: ignore

    def getJson(self) -> dict:
        """
        Load the file written to by the ``LoginManager``.
        """
        with self.savefile.open('r') as f:
            return json.load(f)

    def test_default_address(self):
        self.lm.login(CUBEAddress('https://example.com/api/v1/'), CUBEToken('berry'))
        self.store.set.assert_called_once_with(CUBEAddress('https://example.com/api/v1/'), CUBEToken('berry'))

        content = self.getJson()
        self.assertIn('defaultAddress', content,
                      msg='Login manager did not set the CUBE address as default.')
        self.assertEqual(content['defaultAddress'], 'https://example.com/api/v1/',
                         msg='Default address is incorrect.')

        self.store.get = Mock(return_value='berry')
        self.assertEqual(self.lm.get(), CUBEToken('berry'),
                         msg='Retrieved password for default CUBE address is incorrect.')

        self.lm.logout()
        self.store.clear.assert_called_once_with('https://example.com/api/v1/')
        self.assertNotIn('defaultAddress', self.getJson(),
                         msg='Default address not removed after logout.')



import abc
from dataclasses import dataclass
import requests


@dataclass(frozen=True)
class ConnectedResource(abc.ABC):
    """
    A ``ConnectedResource`` has a :class:`requests.Session` for
    communicating with a *ChRIS* backend.

    CONSTRAINT: ``s`` must have HTTP headers for authentication
                with CUBE and also the header ``Accept: application/json``
    """
    s: requests.Session



from chris.cube.resource.connected_resource import ConnectedResource
from chris.cube.resource.cube_resource import CUBEResource
from chris.cube.resource.templated_resource import ResourceWithTemplate



import json
from typing import Generator, Any, TypedDict, Callable, TypeVar, List, Dict

import requests

from chris.cube.resource import CUBEResource
from chris.types import CUBEUrl

import logging

logger = logging.getLogger(__name__)


class UnrecognizedResponseException(Exception):
    pass


class TooMuchPaginationException(Exception):
    pass


T = TypeVar('T', bound=CUBEResource)


class JSONPaginatedResponse(TypedDict):
    count: int
    next: CUBEUrl
    previous: CUBEUrl
    results: List[Dict[str, Any]]


def fetch_paginated_objects(s: requests.Session,
                            url: CUBEUrl,
                            constructor=Callable[..., T],
                            max_requests=100
                            ) -> Generator[T, None, None]:
    for d in fetch_paginated_raw(s, url, max_requests):
        yield constructor(s=s, **d)


def fetch_paginated_raw(s: requests.Session,
                        url: CUBEUrl, max_requests: int
                        ) -> Generator[Dict[str, any], None, None]:
    """
    Produce all values from a paginated endpoint.

    :param s: session
    :param url: the paginated URI, optionally ending
                with the query-string ``?limit=N&offset=N&``
    :param max_requests: a quota on the number of requests
                         a call to this method may make in total
    """
    if max_requests <= 0:
        raise TooMuchPaginationException()

    logger.debug('%s', url)
    res = s.get(url)  # TODO pass qs params separately?
    res.raise_for_status()
    data = res.json()

    yield from __get_results_from(url, data)
    if data['next']:
        yield from fetch_paginated_raw(s, data['next'], max_requests - 1)


__PaginatedResponseKeys = frozenset(JSONPaginatedResponse.__annotations__)


# TODO in Python 3.10, we should use TypeGuard
# https://docs.python.org/3.10/library/typing.html#typing.TypeGuard
def __get_results_from(url: CUBEUrl, data: Any) -> List[Dict[str, Any]]:
    """
    Check that the response from a paginated endpoint is well-formed,
    and return the results.
    """
    if not isinstance(data, dict):
        logging.debug('Invalid response from %s\n'
                      'Was not parsed correctly into a dict.\n'
                      '%s',
                      url,
                      json.dumps(data, indent=4))
        raise UnrecognizedResponseException(f'Response from {url} is invalid.')

    if __PaginatedResponseKeys > frozenset(data.keys()):
        logging.debug('Invalid response from %s\n'
                      'dict keys did not match: %s\n'
                      '%s',
                      url,
                      str(__PaginatedResponseKeys),
                      json.dumps(data, indent=4))
        raise UnrecognizedResponseException(f'Response from {url} is invalid.')

    return data['results']



from pathlib import Path
import typer
from caw.commands.store import app, build_client
from caw.movedata import download as cube_download

from chris.types import CUBEUrl
from chris.cube.pagination import UnrecognizedResponseException

from caw.constants import DEFAULT_ADDRESS
from caw.commands.helpers import api_address


@app.command()
def download(
        threads: int = typer.Option(4, '--threads', '-t', help='Number of concurrent downloads.'),
        url: str = typer.Argument(..., help='ChRIS files API resource URL'),
        destination: Path = typer.Argument(..., help='Location on host where to save downloaded files.')
):
    """
    Download everything from a ChRIS url.
    """
    base_address = api_address(CUBEUrl(url))
    if base_address != build_client.address:
        if build_client.address != DEFAULT_ADDRESS:
            given_address = typer.style(f'--address={build_client.address}',
                                        fg=typer.colors.GREEN, bold=True)
            download_url = typer.style(url, fg=typer.colors.YELLOW, bold=True)
            typer.echo(f'Given {given_address} is different '
                       f'from download URL: {download_url}', err=True)
            raise typer.Abort()
        build_client.address = base_address

    client = build_client()

    try:
        cube_download(client=client, url=url, destination=destination, threads=threads)
    except UnrecognizedResponseException as e:  # TODO different error please
        typer.secho(str(e), fg=typer.colors.RED, err=True)
        raise typer.Abort()



"""
Badly implemented --- calls ``typer.Abort`` directly, because
I am so done with this.

I want to work on chrs!
"""

import typer
from chris.types import CUBEAddress, CUBEUrl


def api_address(s: CUBEUrl) -> CUBEAddress:
    i = s.find('/api/v1/')
    if i == -1:
        typer.secho(f'Invalid URL: {s}', color=typer.colors.RED, err=True)
        raise typer.Abort()
    return CUBEAddress(s[:i + 8])


assert api_address(CUBEUrl('http://localhost:8080/api/v1/files/')) == CUBEAddress('http://localhost:8080/api/v1/')
assert api_address(CUBEUrl('https://example.com/api/v1/uploadedfiles/')) == CUBEAddress('https://example.com/api/v1/')
assert api_address(CUBEUrl('https://example.com/api/v1/')) == CUBEAddress('https://example.com/api/v1/')



"""
This file eases usage of the Python chris.client and Typer.
"""
import os
from typing import Optional, Tuple

import typer

from chris.cube.pipeline import Pipeline
from chris.client import ChrisClient
from chris.errors import ChrisIncorrectLoginError, PipelineNotFoundError

from chris.types import CUBEAddress, CUBEToken, CUBEUsername, CUBEPassword
from caw.constants import DEFAULT_ADDRESS, DEFAULT_USERNAME, DEFAULT_PASSWORD

from caw.login.manager import LoginManager

import logging
logger = logging.getLogger(__name__)


class FriendlyClient(ChrisClient):
    """
    A ``ChrisClient`` which shows (less helpful) error messages instead of exceptions.
    """
    def get_pipeline(self, name: str) -> Pipeline:
        try:
            return super().get_pipeline(name)
        except PipelineNotFoundError:
            typer.secho(f'Pipeline not found: "{name}"', fg=typer.colors.RED, err=True)
            raise typer.Abort()


class ClientBuilder:
    """
    A workaround so that the ChrisClient object's constructor, which attempts to use the login
    credentials, is called by the subcommand instead of the main callback.

    This is necessary to support the ``caw login`` subcommand.
    """
    def __init__(self, login_manager: LoginManager):
        self.login_manager = login_manager
        self.address: Optional[CUBEAddress] = None
        self.username: Optional[CUBEUsername] = None
        self.password: Optional[CUBEPassword] = None

    def _determine_address(self, given_address: Optional[CUBEAddress]) -> CUBEAddress:
        if given_address:
            return given_address
        saved_address = self.login_manager.get_default_address()
        if saved_address:
            return saved_address
        return DEFAULT_ADDRESS

    def _determine_login_method(self, address: CUBEAddress) \
            -> Tuple[Optional[CUBEToken], Optional[CUBEUsername], Optional[CUBEPassword]]:
        """
        Determines whether to use token auth, or username+password auth.

        1. If user explicitly specifies a username and password, then use them.
        2. If neither username nor password given, then check if they're logged in and get the token.
        3. If user is not logged in and does not give username nor password, then use default username and password.

        :param address: address of ChRIS to log into
        :return: token, username, password
        """
        if self.username and not self.password:
            typer.echo('Given a username but no password. You must supply both.', err=True)
            raise typer.Abort()
        if self.password and not self.username:
            typer.echo('Given a password but no username. You must supply both.', err=True)
            raise typer.Abort()

        if self.username and self.password:
            return None, self.username, self.password

        token = self.login_manager.get(address)
        if token:
            return token, None, None

        return None, DEFAULT_USERNAME, DEFAULT_PASSWORD

    @staticmethod
    def _create_client_with(address: CUBEAddress, token: Optional[CUBEToken],
                            username: Optional[CUBEUsername], password: Optional[CUBEPassword]) -> FriendlyClient:
        """
        Use token authentication if given a token. Otherwise, use username and password.
        """
        if token:
            return FriendlyClient(address=address, token=token)
        return FriendlyClient.from_login(address=address, username=username, password=password)

    def __call__(self) -> FriendlyClient:
        """
        Authenticate with ChRIS and construct the client object.

        In general, it works like this:

        1. First try to use user-specified values.
        2. For any value not specified, check to see if they're logged in and use their existing session.
        3. As a last resort, use defaults (e.g. chris:chris1234 on http://localhost:8000/api/v1/)

        :return: client object
        """
        address = self._determine_address(self.address)
        token, username, password = self._determine_login_method(address)
        self._print_debug(address, token, username, password)

        try:
            return self._create_client_with(address, token, username, password)
        except ChrisIncorrectLoginError as e:
            self._handle_incorrect_login(e, address, token)
        # except Exception as e:
        #     typer.secho('Unknown connection error', fg=typer.colors.RED, err=True)
        #     typer.echo(str(e), err=True)
        #     raise typer.Abort()

    def _handle_incorrect_login(self, e: ChrisIncorrectLoginError, address: CUBEAddress,
                                token: Optional[CUBEToken] = None):
        typer.secho(f'Authentication failed for {address}.', err=True)
        typer.secho(str(e), err=True)

        if token:
            typer.secho(f'warning: removing saved login for {address}.\n'
                        f'To login again, run\n'
                        f"\n\tcaw --address '{address}' login", err=True)
            self.login_manager.logout(address)
        raise typer.Abort()

    @staticmethod
    def _print_debug(address: CUBEAddress, token: Optional[CUBEToken],
                     username: Optional[CUBEUsername], password: Optional[CUBEPassword]):
        if token:
            logger.debug('HTTP token: "%s"', token)
            return
        if password == DEFAULT_PASSWORD and 'CHRIS_TESTING' not in os.environ:
            typer.secho('Using defaults (set CHRIS_TESTING=y to suppress this message): '
                        f'{address}  {username}:{password}', dim=True, err=True)
            return
        logger.debug('Using HTTP basic auth with given username and password.')
        logger.debug('address=%s username=%s', address, username)



import os
import unittest
from tempfile import NamedTemporaryFile
from pathlib import Path

from chris.client import ChrisClient
from caw.movedata import upload


example_data1 = r"""
Remember, Grasshopper, falling down 1000 stairs begins by tripping over
the first one.
                -- Confusion
"""

example_data2 = r"""
The steady state of disks is full.
                -- Ken Thompson
"""


class TestUpload(unittest.TestCase):

    client = ChrisClient.from_login(
        address='http://localhost:8000/api/v1/',
        username='chris',
        password='chris1234'
    )

    def setUp(self) -> None:
        with NamedTemporaryFile(mode='w', newline='\n', encoding='utf-8', suffix='.txt',  delete=False) as tf1:
            tf1.write(example_data1)
        with NamedTemporaryFile(mode='w', newline='\n', encoding='utf-8', suffix='.txt',  delete=False) as tf2:
            tf2.write(example_data2)
        self.file1 = tf1
        self.file2 = tf2

    def tearDown(self) -> None:
        os.unlink(self.file1.name)
        os.unlink(self.file2.name)

    def test_upload(self):
        upload(
            client=self.client,
            files=[Path(self.file1.name), Path(self.file2.name)],
            parent_folder='caw/test_upload',
            upload_threads=2
        )

        results = self.client.search_uploadedfiles(fname='chris/uploads/caw/test_upload/')
        fnames = [os.path.basename(result.fname) for result in results]
        self.assertIn(os.path.basename(self.file1.name), fnames)
        self.assertIn(os.path.basename(self.file2.name), fnames)



from os import path
from setuptools import setup, find_packages

with open(path.join(path.dirname(path.abspath(__file__)), 'README.md')) as f:
    readme = f.read()

setup(
    name='caw',
    version='0.4.1',
    packages=find_packages(exclude=('chris.tests.*', 'caw.tests.*')),
    url='https://github.com/FNNDSC/caw',
    license='MIT',
    author='Jennings Zhang',
    author_email='Jennings.Zhang@childrens.harvard.edu',
    description='ChRIS Automatic Workflows',
    long_description=readme,
    long_description_content_type='text/markdown',
    python_requires='>=3.8.2',
    install_requires=['requests', 'typer', 'shellingham', 'packaging'],
    entry_points={
        'console_scripts': [
            'caw = caw.__main__:app'
            ]
        },
    classifiers=[
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ]
)



import abc
from typing import TypedDict, List
from dataclasses import dataclass, InitVar
from chris.cube.resource.cube_resource import CUBEResource


class CollectionEntry(TypedDict):
    name: str
    value: str


class Template(TypedDict):
    data: List[CollectionEntry]


@dataclass(frozen=True)
class ResourceWithTemplate(CUBEResource, abc.ABC):
    template: InitVar[Template]

    def __post_init__(self, template: Template):
        pass



import typer
from caw.commands.store import app, build_client


@app.command()
def search(name: str = typer.Argument('', help='name of pipeline to search for')):
    """
    Search for pipelines that are saved in ChRIS.
    """
    client = build_client()
    for search_result in client.search_pipelines(name):
        typer.echo(f'{search_result.url:<60}{typer.style(search_result.name, bold=True)}')



import pytest
from pytest_mock import MockerFixture
from unittest.mock import Mock, call

from chris.types import CUBEUrl
from chris.cube.pagination import fetch_paginated_raw

from chris.tests.mocks.session import mock_session
from chris.tests.mocks.data.pagination import responses


@pytest.fixture
def session(mocker: MockerFixture) -> Mock:
    return mock_session(mocker, responses)


def test_pagination(session: Mock):
    url = CUBEUrl('https://example.com/api/v1/something/')
    all_things = list(fetch_paginated_raw(session, url, max_requests=10))
    assert all_things == [
        {'id': 1}, {'id': 2}, {'id': 3},
        {'id': 4}, {'id': 5}, {'id': 6},
        {'id': 7}, {'id': 8}
    ]
    session.get.assert_has_calls([
        call('https://example.com/api/v1/something/'),
        call('https://example.com/api/v1/something/?limit=3&offset=3'),
        call('https://example.com/api/v1/something/?limit=3&offset=6')
    ])
